/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*!
  \file BA.h
  \brief Contains ba (BasinArray), an N-dimensional distributed array class, based upon MPI-2.
 
Basin array class allows to define a N-dim array, automatically 
distribute its elements among n processes in a distributed memory machine 
and access them "directly" as in a shared-memory machine.
Non-contiguous data can be accessed and assigned.
This class has been written for the data management of the BASIN project.
It requires MPI-2.
*/

#include "ba_trait.h"
#include <basin_base.h>

#include "fparser.hh"
#include "promote-old.h"

using namespace std;

template <typename T> 
/** \class ba BA.h "BA.h"
    \brief Distributed N-dimensional array class
    
    ba or BasinArray is a distributed N-dim array.  Different processes can
    access data "directly" and non-contiguous data can be worked on with
    relative ease.  This class is based largely on the one-sided operations
    in MPI-2.
*/

class ba
{
private:
  /* 
     dimensions           vector<int> of dimensions
     number_of_dimensions number of dimensions of the array.
     dim_v:               total number of elements in the BA
     loc_dim_v:           dimension of the local vector (portion of vector in a
                          processor)
     n_proc:              number of processors
     size_T:              size of datatype
     rank:                rank of the process
     map_u:               array of upper limits (in term of global index) 
                          of each processor.
     last_process_reserv: number of extra elements allowed in the last
                          process (beyond dim_v)
     last_index_put:      index to keep track of how many elements have been 
                          put in the reserve memory.
			  dim_v+last_index_put=current number of elements 
			  in the BA
  */
  /// Total number of elements contained within the array
  int dim_v;
  /// Dimension of the local vector
  int loc_dim_v;
  /// The actual number of elements in the array
  int number_of_dimensions;
  /// Number of processors
  int n_proc;
  /// Size of the datatype
  int size_T;
  /// This processor's rank
  int rank;
  /// Array of upper limits (in terms of the global index) for each processor
  int *map_u;
  /// Number of extra "reserve" elements allowed beyond dim_v
  int last_process_reserve;
  /// Index to keep track of elements put into reserve memroy
  int last_index_put;
  /// The number of elements in each dimension
  vector<int> dimensions;
  /// Comm we are running the ba in
  MPI_Comm comm;
  /// The MPI_Datatype the ba is holding
  MPI_Datatype data_T;
  /// Window for the ba
  MPI_Win win;
  /// Our actual data
  T *g;
  
public:
  typedef T td;

  /*!Constructor */
  ba();
  /*!Copy Constructor */
  ba(const ba<T> &a);
  /*!Copy Constructor */
  /* made explicit to avoid implicit conversion in operations*/
  explicit ba(const T a);

  ~ba(void);

  /*!Set the dimensions of the array, the MPI_comm where the BA will
     operate, the reserve of "rows" to be allocated in the last process. 
     This function sets the map for data distribution by evenly distributing
     the "number_of_rows" (number of elements of the left-most index)      
     and create the BA.\\
     {\b NOTE: this function must be called after MPI_Init and by all processes.}
     @param vec_d STL vector(size=number_of_dimensions of the basin_array) 
      of integers specifying the size the basin_array in each dimension.
     (e.g. for a 2-d basin_array 
      vec_d[0]=number_of_rows, vec_d[1]=number_of_columns) 
     @param c MPI_Comm where the basin array will operate
     @param reserv number of extra reserve "rows" to be allocated (beyond the 
     vec_d[0] rows)
     
  */
  int ba_create(vector<int> vec_d, MPI_Comm c,int reserv);
  int ba_internal_create(vector<int> vec_d, MPI_Comm c,int reserv);
  int ba_create2();

  /*! Free the memory and the window (see MPI-2 manual for what a window is)*/
  void ba_free();

  /*! Get the global lower and upper index for the rows of
     data assigned to process with rank=proc_rank 
     For example for
     10x10x10 grid distributed among three processes, this function 
     will return  
     ilo=0,ihi=3 for proc_rank=0
     ilo=4,ihi=6 for proc_rank=1
     ilo=7,ihi=9 for proc_rank=2     
     @param proc_rank: rank of the process 
     @param int*ilo: (lower limit index for the "rows" owned by process proc_rank) 
     @param int*ihi: (upper limit index for the "rows" owned by process proc_rank) 
*/
  void ba_distribution(int proc_rank,int *ilo, int *ihi) const;

  vector<int> ba_distribution(int proc_rank) const;



  /*! Get the global lower and upper 1d index  of the
     data assigned to process with rank=proc_rank. For example for
     10x10x10 grid distributed among three processes, this function 
     will return  
     ilo=0,ihi=399 for proc_rank=0
     ilo=400,ihi=699 for proc_rank=1
     ilo=700,ihi=999 for proc_rank=2
     @param proc_rank: rank of the process 
     @param int*ilo: (lower limit index owned by process proc_rank) 
     @param int*ihi: (upper limit index owned by process proc_rank) 
*/
  void ba_distribution_index(int proc_rank,int *ilo, int *ihi) const;
  vector<int> ba_distribution_index(int proc_rank) const;

  /*! Put ptr elements in the basin array BA[lo]....Ba[hi].

@param lo STL vector<int> of size=dimensions_of_the_array with lower
limits of the array region where data are to be put. 

@param hi STL vector<int> of size=dimensions_of_the_array with
indexes specifying the lower limits in each dimension of 
the basin_array region where data are to be put.
@doc \flushleft Example:

2-d basin_array A

To put 4 values contained in ptr[0]...ptr[3] in 
A[1][3], A[1][4], A[2][3], A[2][4],

lo[0]=1,hi[0]=2

lo[1]=3,hi[1]=4

Note that lo and hi are global indexes, you do not have to worry about
which process actually owns the elements of the array to which you are
assigning the ptr values*/ 
  int ba_putF(const vector<int> lo,const vector<int> hi,T *ptr);
  int ba_put(const vector<int> lo,const vector<int> hi,T *ptr);
  //template <typename U>
  //int ba_put(const vector<int> lo,const vector<int> hi,T *ptr,1);
  // int ba_put(const vector<int> lo,const vector<int> hi,T *ptr);
  // Instead of sending ordered pairs as corners, send the actual
  // index limits.  Does not check for dimensionality.
  int ba_put(const int, const int, T* ptr);

  /*! Add the values of ptr  to the existing values of the BA elements with 
    the given indexes*/
  int ba_add(const vector<int> lo,const vector<int> hi,T *ptr);
  // Instead of sending ordered pairs as corners, send the actual
  // index limits.  Does not check for dimensionality.
  int ba_add(const int, const int, T* ptr);



  /*! Multiply the values of ptr  to the existing values of the BA elements 
    with the given indexes*/
  int ba_multiply(const vector<int> lo,const vector<int> hi,T *ptr);
  // Instead of sending ordered pairs as corners, send the actual
  // index limits.  Does not check for dimensionality.
  int ba_multiply(const int, const int, T* ptr);



  /*! append elements ptr[lo]...ptr[hi] to the end of the BA*/
  int ba_push_back(int numb_elems, T *ptr);

  /*! \flushleft Get elements BA[lo]....Ba[hi] and put them into ptr.

	   @param  lo STL vector<int> of size = dimensions_of_the_array
	   with lower limits of the array region where data are to be
	   gotten from.   
	   @param hi STL vector<int> of size = dimensions_of_the_array
	   containing the indexes specifying the lower limits in each 
	   dimension of the basin_array region where data are to be gotten from. 
	   @doc \flushleft Example:

	   2-d basin array A. To get the 4 values contained  in 
	   A[1][3], A[1][4], A[2][3], A[2][4] and put them in
	   ptr[0]...ptr[3]
 
	   lo[0]=1,hi[0]=2

	   lo[1]=3,hi[1]=4

	   Note that lo and hi are global indexes, you do not have to
	   worry about which process actually owns the elements of the
	   array you are getting (however access to the data owned by
	   the process calling this function will be faster)*/
  int ba_get(const vector<int> lo,const vector<int> hi,T* ptr) const;
  // Instead of sending ordered pairs as corners, send the actual
  // index limits.  Does not check for dimensionality.
  int ba_get(const int, const int, T* ptr) const;
  int ba_get_opt(const int, T* ptr) const;
  int ba_get_opt(const vector<int>, T* ptr) const;


  /*! fill the BA with a given value*/
  int ba_fill(T fv);


  /*! return the total number of elements in a BA*/
  int ba_size() const;


  /*! return a vector with the upper limit indexes of each basin array dimension (e.g. for a 2-d 5x3 basin array it returns a vector whose elements (4,2))*/
  vector<int> ba_get_dims() const;


  /*! return the local number of elements in a BA*/
  int ba_get_loc_dimv() const;


  /*! return a vector with the limits of each basin array dimension (e.g. for a 2-d 5x3 basin array it returns a vector whose elements (5,3))*/
  vector<int> ba_get_dimensions() const;


  MPI_Comm ba_get_comm() const;


  /*! direct access to the local data by a pointer- no copy */
  void ba_get_access(T**ptr) const;
  T* ba_get_access() const;
  T ba_get_access(int i) const;
  void ba_set(int i, T sv);

  /*! operations on ba*/
  int add(ba<T> a,ba<T> b);
  int subtract(ba<T> a,ba<T> b);
  int multiply(ba<T> a,ba<T> b);
  int divide(ba<T> a,ba<T> b);
  int evaluate(ba<T> a,std::string func);
  int evaluate(ba<T> a, ba<T> b,std::string func);
  int evaluate(ba<T> a, ba<T> b,ba<T> c,std::string func);
  T operator [] (const int index);
  // ba<T> operator+ (T y);
  //  ba<T> operator- (T y);
  //  ba<T> operator* (T y);
  //  ba<T> operator/ (T y);
  ba<T> operator^ (T y);
  ba<T> &operator= (const ba<T> &b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote> operator+ (const ba<U> &b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote> operator+ (U b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote> operator- (const ba<U> &b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote> operator- (U b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote_pr> operator* (const ba<U> &b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote_pr> operator* (U b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote> operator/ (const ba<U> &b);
  template <typename U> ba<typename promote_trait<T,U>::T_promote> operator/ (U b);

  template <typename U> ba<int> operator< (const ba<U> &b);
  template <typename U> ba<int> operator< (U b);
  template <typename U> ba<int> operator> (const ba<U> &b);
  template <typename U> ba<int> operator> (U b);
  template <typename U> ba<int> operator>= (const ba<U> &b);
  template <typename U> ba<int> operator>= (U b);
  template <typename U> ba<int> operator<= (const ba<U> &b);
  template <typename U> ba<int> operator<= (U b);
  template <typename U> ba<int> operator== (const ba<U> &b);
  template <typename U> ba<int> operator== (U b);
  template <typename U> ba<int> operator!= (const ba<U> &b);
  template <typename U> ba<int> operator!= (U b);

  template <typename U> ba<int> operator&& (const ba<U> &b);
  template <typename U> ba<int> operator|| (const ba<U> &b);


  /*! get the window of the global array*/
  MPI_Win ba_win();

  /*! Sort function*/
  int ba_Sort(vector<aux_vect<T> > &);


  /* Old version of make_sortable and new version taking an STL vector 
     as an argument instead.
     Both these version are probably not needed anymore 
     (see the new version below) but retained for compatibility.
  */
  vector<aux_vect<T> > make_sortable(T *y,int lo, int hi);
  vector<aux_vect<T> > make_sortable(vector<T> y,int lo, int hi);

  /* 
     The previous version of make_sortable is retained for compatibility
     but since y, lo and hi will always be those of the local values
     of each process, we don't need to pass them. 
  */
  vector<aux_vect<T> > make_sortable();

};

