/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*! \file basin.h
  \brief Contains the standard includes for the project as well as some defines*/
  

/*! \mainpage Beowulf Analysis Symbolic INterface
Parallel computing has become an increasingly important tool for scientists working with large datasets. A ``Beowulf'' cluster utilizes commodity personal computers and off-the-shelf switches to achieve very cost-effective performance compared to other supercomputing architectures. A generic problem in computational physics is that vastly different physical problems frequently require very similar numerical treatments; however, individual research groups have no easy way to effectively share analysis tools, and thus must constantly re-invent (or at least re-code) the wheel.  Certain operations (such as performing coordinate transformations, binning data, and calculating statistical moments of a distribution) are ubiquitous enough that a common interface to perform these manipulations would be of great value to the community.  This is the motivation for the development of the Beowulf Analytical Symbolic INterface application (BASIN), a five-year, NSF-funded program to develop a comprehensive suite of parallelized analysis tools for Astrophysics.

BASIN will include an symbolic interface similar in structure to those found in Maple or Mathematica. It will be able to compute parallelized functions, calling on modified versions of open source single-computer based software as well as newly developed routines. It will also have the ability to display visualizations of the data in novel ways, such as the time evolution of many massive bodies in a space. BASIN will be open source software so it will be accessible and modifiable for many research interests.

 */



//***************************************************************
//
//   HEADER INFORMATION
//
//
#ifndef BASIN_BASE_H
#define BASIN_BASE_H

enum data_type {NONE, INT, FLOAT, DOUBLE, STRING, CHAR, GRID, LIST};

//C++ STL Includes
#include "mpi.h"
#include <string>
#include <list>
#include <map> 
#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <complex>


//C Includes
#include <cstdio>
#include <cmath>
#include <cfloat>

//serialization header files

/** @name Defined Metric Types
    The following are currently defined metrics to be used within BASIN.*/
//@{
///Cartesian coordinates
#define CARTESIAN 0
///Spherical coordinates
#define SPHERICAL 1
///Cylindrical coordinates
#define CYLINDRICAL 2
///Some other coordinate system
#define OTHER -1
//@}
/** @name Defined Boundary Types
    The following are currently defined boundary types to be used within BASIN.*/
//@{
/// Isolated Boundary
#define ISOLATED 0
/// Periodic Boundary
#define PERIODIC 1
//@}

/** @name Defined Space Type
    The following are currently defined space types to be used within BASIN.*/
//@{
/// Real space
#define REAL 0
/// Fourier Hermite
#define FOURIER_HERMITE 1
/// Fourier general
#define FOURIER_GENERAL 2
//@}

// Local junk
// #include "gen_struct.h"

#include "fftw_includes.h"

#include "limits.h"
#include "float.h"

using namespace std;
typedef std::complex<double> dcomplex;




#define BASIN_nxDefault vector<int>(3,10)
//vector<int> BASIN_nxDefault(3,10);
#define BASIN_posDefault vector<double>(1,DBL_MAX)
//vector<float> BASIN_posDefault(1,FLT_MAX);
#define BASIN_selDefault vector<int>(3,1)

#include "basin_config.h"

extern MPI_Comm BASIN_COMM;

#endif

//#ifdef MPE
//#include "mpe_flags.h"
//#include <mpe.h>
//#endif


